/**
 * <!--------------------------------------------------------------------------
 * This file is part of InVivoCrowdingSimulator.  Please visit
 * http://fun.bio.keio.ac.jp/software/invivosim/ for more
 * information about InVivoCrowdingSimulator and its latest version.
 *
 * Copyright (C) 2011-2012 by the Keio University, Yokohama, Japan
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.  A copy of the license agreement is provided
 * in the file named "LICENSE.txt" included with this software distribution.
 * ---------------------------------------------------------------------- -->*/
/*
 * File Name:
 *    dump_space.c
 *
 * Description:
 *    3D lattice-based simulation for studying the dynamics of particles in the crowded environment.
 *    This outputs reaction space (crowded environment) as a text file (space.txt).
 *    You can visualize the reaction space by executing a script 'plot3d.py'.
 *
 * Last modified: Fri, 04 May 2012 22:40:39 +0900
 *
 * GCC Version:
 *    4.2 or later is required (if you want to enable OpenMP).
 *    $ gcc --version
 *
 * Compile:
 *    (linux) $ gcc44 dump_space.c -std=c99 -fopenmp -lgomp -lgsl -lgslcblas
 *    (mac)   $ gcc dump_space.c -std=c99 -fopenmp -lgsl
 *
 * Runtime Arguments:
 *    arg.1: lattice size (> 0)
 *    arg.2: obstacle rate (>= 0.0 and < 1.0)
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <omp.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_types.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_inline.h>

/*** type declaration ***/
typedef unsigned char       uchar;
typedef unsigned short      ushort;
typedef unsigned int        uint;
typedef unsigned long       ulong;
typedef long long           llong;
typedef unsigned long long  ullong;

#define FIELD_FREE      true
#define FIELD_OBSTACLE  false


/*** grobal variables ***/
struct config {
    // runtime arguments
    uint    latticeSize;
    uint    maxSteps;
    uint    trialNum;
    double  obstacleRate;

    // pre-caliculated value
    uint    volume;
    uint    square;
    uint    obstacleNum;
} config;

#ifdef _OPENMP
gsl_rng **rng;
#else
gsl_rng *rng;
#endif

ullong  *results;


/*** configurations */
void initConfig (char **argv)
{
    config.latticeSize  = atoi(argv[1]); 
    config.maxSteps     = 1;
    config.obstacleRate = atof(argv[2]);
    config.trialNum     = 1;
    config.volume       = config.latticeSize * config.latticeSize * config.latticeSize;
    config.square       = config.latticeSize * config.latticeSize;
    config.obstacleNum  = config.volume * config.obstacleRate;
}

bool configValidate ()
{
    if (config.latticeSize <= 0 || 
            config.volume <= config.obstacleNum) {
        return false;
    } 

    return true;
}


void configDump()
{
    fprintf(stderr, "Lattice Size:  %d\n", config.latticeSize);
    fprintf(stderr, "Max Steps:     %d\n", config.maxSteps);
    fprintf(stderr, "Obstacle Rate: %lf\n", config.obstacleRate);
    fprintf(stderr, "Trial Number:  %d\n", config.trialNum);
}

void usage ()
{
    fprintf(stderr, "Usage: [lattice size] [obstacle rate]\n");
    exit(EXIT_FAILURE);
}


/*** utility ***/
static inline uint xyzToIndex (uint *coord)
{
    uint size = config.latticeSize;
    return (coord[0] * size + coord[1]) * size + coord[2];
}

static inline void indexToXyz (uint index, uint *coord)
{
    uint size   = config.latticeSize;
    uint square = config.square;
    coord[0]    = index / square;
    coord[1]    = (index / size) % size;
    coord[2]    = index % size;
}

void* reserveMemory (uint size)
{
    void* ptr = (void*)malloc(size);
    if (ptr == NULL) {
        fprintf(stderr, "cannot reserve memory\n");
        exit(EXIT_FAILURE);
    }
    return ptr;
}

static inline ullong calMsd (int *dist)
{
    return (llong)dist[0] * (llong)dist[0] + (llong)dist[1] * (llong)dist[1] + (llong)dist[2] * (llong)dist[2]; 
}


/*** random number generator ***/
gsl_rng* createRng (ulong seed)
{
    gsl_rng* rng;
    gsl_rng_env_setup();
    rng = gsl_rng_alloc(gsl_rng_default);
    gsl_rng_set(rng, seed);
    return rng;
}

void initRng ()
{
    ulong seed = time(NULL);

    #ifdef _OPENMP
    rng = reserveMemory(sizeof(gsl_rng) * omp_get_max_threads());
    for (int i = 0; i < omp_get_max_threads(); ++i) {
        rng[i] = createRng(seed + i); 
    } 
    #else
    rng = createRng(seed); 
    #endif
}

static inline gsl_rng* getRandGen ()
{
    gsl_rng* randGen;

    #ifdef _OPENMP
    randGen = rng[omp_get_thread_num()];
    #else
    randGen = rng; 
    #endif

    return randGen;
}

static inline uint createRandNum (gsl_rng *randGen, uint max)
{
    return max * gsl_rng_uniform(randGen);
}

void freeRng()
{
    #ifdef _OPENMP
    for (int i = 0; i < omp_get_max_threads(); ++i) {
        gsl_rng_free(rng[i]);
    }
    free(rng);
    #else
    gsl_rng_free(rng);
    #endif
}


/*** simulation ***/
/* initialize */
void initSpace (bool *space)
{
    uint volume, obstacleNum, index, replaceIndex;
    bool *boolIndex, boolVal;
    gsl_rng *randGen;

    volume      = config.volume;
    obstacleNum = config.obstacleNum;
    index       = volume - 1;
    boolIndex   = space + volume - 1;
    randGen     = getRandGen();

    memset(space, FIELD_OBSTACLE, sizeof(bool) * obstacleNum);
    memset(space + obstacleNum, FIELD_FREE, sizeof(bool) * (volume - obstacleNum));

    while (index != 1) {
        replaceIndex = (index + 1) * gsl_rng_uniform(randGen);
        boolVal = space[replaceIndex];
        space[replaceIndex] = *boolIndex;
        *boolIndex = boolVal;
        --boolIndex;
        --index;
    }
}

uint initIndex (bool *space)
{
    uint volume, index;
    gsl_rng *randGen;

    volume  = config.volume;
    randGen = getRandGen();

    while (1) {
        index = volume * gsl_rng_uniform(randGen);
        if (space[index]) break; 
    }

    return index;
}


/* move molecules */
static inline uint getNextCoord (int coord)
{
    uint size = config.latticeSize;
    if (coord == -1) {
        coord += size;
    } else if (coord == size) {
        coord = 0;
    } 

    return coord;
}

static inline bool checkMobility (bool *space, uint *orgCoord, uint *tmpCoord)
{
    uint index, axis, size;
    int move;

    size    = config.latticeSize;
    index   = 0;

    for (int i = 0; i < 3; ++i) {
        tmpCoord[i] = orgCoord[i];
    }

    while (index != 6) {
        axis = index / 2;
        move = 2 * (index % 2) - 1;
        tmpCoord[axis] = getNextCoord(orgCoord[axis] + move);

        if (space[xyzToIndex(tmpCoord)]) break;

        tmpCoord[axis] = orgCoord[axis];
        ++index;
    }

    if (index == 6) {
        return false;
    } else {
        return true;
    }
}

void run (bool *space, uint *currentCoord, uint *nextCoord, int *dist)
{
    // variables
    gsl_rng *randGen;
    uint maxSteps, size, axis, randNum;
    int move; 

    // initialize
    maxSteps    = config.maxSteps;
    size        = config.latticeSize;
    randGen     = getRandGen();
    initSpace(space);
    indexToXyz(initIndex(space), currentCoord);

    // dump space
    uint *nor_pos  = reserveMemory(sizeof(uint) * 3);
    FILE *fp = fopen("space.txt", "w");
    for (int i = 0; i < config.volume; ++i) {
      if (*(space+i) == FIELD_OBSTACLE) {
        indexToXyz(i, nor_pos);
        fprintf(fp, "%d\t%d\t%d\n", nor_pos[0], nor_pos[1], nor_pos[2]);
      }
    }
    fclose(fp);
}

void runSim ()
{
    uint trialNum   = config.trialNum;
    initRng();

    // parallelization
    #ifdef _OPENMP
    omp_set_num_threads(omp_get_max_threads());
    #endif

    #pragma omp parallel
    {
        bool *space         = reserveMemory(sizeof(bool) * config.volume);
        uint *currentCoord  = reserveMemory(sizeof(uint) * 3);
        uint *nextCoord     = reserveMemory(sizeof(uint) * 3);
        int  *dist          = reserveMemory(sizeof(int)  * 3);

        #pragma omp for
        for (int i = 0; i < trialNum; ++i) {
            fprintf(stderr, "Running Trial No.%d\n", i + 1);
            run (space, currentCoord, nextCoord, dist); 
        }

        free(space);
        free(currentCoord);
        free(nextCoord);
        free(dist);
    }

    freeRng();
}


/*** results ***/
void resultsInit()
{
    uint maxSteps = config.maxSteps;
    results = (ullong *)malloc(sizeof(ullong) * maxSteps);
    memset(results, 0, sizeof(ullong) * maxSteps);
}

void resultsDump()
{
    for (uint i = 0; i < config.maxSteps; ++i) {
        fprintf(stdout, "%d %lf\n", i, (double)results[i] / (double)config.trialNum);
    }
}


/*** main function ***/
int main (int argc, char **argv)
{
    if (argc != 3) {
        usage(); 
    }

    // configurations
    fprintf(stderr, "### Configurations ###\n");
    initConfig(argv);
    if (!configValidate()) usage();
    configDump();

    // simulation
    fprintf(stderr, "\n### Simualtion started ###\n");
    resultsInit();
    runSim();

    // result
    fprintf(stderr, "\n### Dump reaction space as space.txt ###\n");

    free(results);

    return 0;
}
