/**
 * <!--------------------------------------------------------------------------
 * This file is part of InVivoCrowdingSimulator.  Please visit
 * http://fun.bio.keio.ac.jp/software/invivosim/ for more
 * information about InVivoCrowdingSimulator and its latest version.
 *
 * Copyright (C) 2011-2012 by the Keio University, Yokohama, Japan
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.  A copy of the license agreement is provided
 * in the file named "LICENSE.txt" included with this software distribution.
 * ---------------------------------------------------------------------- -->*/
/*
 * File Name:
 *    msd_in_move_sphere.c
 *
 * Description:
 *    3D lattice-based simulation for studying the dynamics of sphere particles in the crowded environment.
 *    This outputs MSD (Mean Square Displacement) corresponding to the time.
 *    This simulation use "Blind Ant model" for Brownian motion.
 *
 *    sphere particles move per interval
 *
 * Last modified: Fri, 04 May 2012 22:16:24 +0900
 *
 * GCC Version:
 *    4.2 or later is required (if you want to enable OpenMP).
 *    $ gcc --version
 *
 * Compile:
 *    (linux) $ gcc44 msd.c -std=c99 -fopenmp -lgomp -lgsl -lgslcblas
 *    (mac)   $ gcc msd.c -std=c99 -fopenmp -lgsl
 *
 * Runtime Arguments:
 *    arg.1: lattice size (> 0)
 *    arg.2: max steps (> 0)
 *    arg 3: obstacle radius
 *    arg.4: obstacle rate (>= 0.0 and < 1.0)
 *    arg.5: obstalce moving interval
 *    arg.6: trial number (> 0)
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <limits.h>
#include <omp.h>
#include <gsl/gsl_rng.h>

/*** type declaration ***/
typedef unsigned char       uchar;
typedef unsigned short      ushort;
typedef unsigned long       ulong;
typedef long long           llong;
typedef unsigned long long  ullong;

#define FIELD_FREE      true
#define FIELD_OBSTACLE  false


/*** grobal variables ***/
typedef struct stCoord {
  int     x;
  int     y;
  int     z;
} stCoord;

#ifdef _OPENMP
gsl_rng **rng;
#else
gsl_rng *rng;
#endif

ullong  *results;

struct config {
  // runtime arguments
  int    latticeSize;
  int    maxSteps;
  int    trialNum;
  int    obstacleRadius;
  int    obstacleMovingInterval;
  int    obstacleNum;
  double  obstacleRate;

  // pre-caliculated value
  int    volume;
  int    square;

  // obstacle
  int    sphereNum;
  int    sphereSize;
  int    excludeVolumeSize;
  stCoord *sphere;
  stCoord *excludeVolume;
} config;

/*** utility ***/
static inline int xyzToIndex (int *coord)
{
  int size = config.latticeSize;
  return (coord[0] * size + coord[1]) * size + coord[2];
}

static inline void indexToXyz (int index, int *coord)
{
  int size   = config.latticeSize;
  int square = config.square;
  coord[0]    = index / square;
  coord[1]    = (index / size) % size;
  coord[2]    = index % size;
}

void* reserveMemory (int size)
{
  void* ptr = (void*)malloc(size);
  if (ptr == NULL) {
    fprintf(stderr, "cannot reserve memory\n");
    exit(EXIT_FAILURE);
  }
  return ptr;
}

static inline ullong calMsd (int *dist)
{
  return (llong)dist[0] * (llong)dist[0] + (llong)dist[1] * (llong)dist[1] + (llong)dist[2] * (llong)dist[2]; 
}

/*** move molecules ***/
bool findElement (stCoord *array, int size, int x, int y, int z)
{
  for (int i = 0; i < size; ++i) {
    if (array[i].x == x && array[i].y == y && array[i].z == z) {
      return true; 
    } 
  } 
  return false;
}

void adoptPeriodicBoundaryCondition (int* tmpCoord)
{
  const int size = (int)config.latticeSize;
  for (int i = 0; i < 3; ++i) {
    while (tmpCoord[i] >= size) tmpCoord[i] -= size; 
    while (tmpCoord[i] < 0) tmpCoord[i] += size; 
  }
}

static inline int getNextCoord (int coord)
{
  int size = config.latticeSize;
  if (coord == -1) {
    coord += size;
  } else if (coord == size) {
    coord = 0;
  } 

  return coord;
}

static inline bool checkMobility (bool *space, int *orgCoord, int *neighborCoord)
{
  int index, axis, size;
  int move;

  size    = config.latticeSize;
  index   = 0;

  for (int i = 0; i < 3; ++i) {
    neighborCoord[i] = orgCoord[i];
  }

  while (index != 6) {
    axis = index / 2;
    move = 2 * (index % 2) - 1;
    neighborCoord[axis] = getNextCoord(orgCoord[axis] + move);

    if (space[xyzToIndex(neighborCoord)]) break;

    neighborCoord[axis] = orgCoord[axis];
    ++index;
  }

  if (index == 6) {
    return false;
  } else {
    return true;
  }
}


/*** configurations */
void initConfig (char **argv)
{
  config.latticeSize              = atoi(argv[1]);
  config.maxSteps                 = atoi(argv[2]);
  config.obstacleRadius           = atoi(argv[3]);
  config.obstacleRate             = atof(argv[4]);
  config.obstacleMovingInterval   = atoi(argv[5]);
  config.trialNum                 = atoi(argv[6]);
  config.volume                   = config.latticeSize * config.latticeSize * config.latticeSize;
  config.square                   = config.latticeSize * config.latticeSize;
  config.obstacleNum              = config.volume * config.obstacleRate;

  // obstacle
  const int radius        = config.obstacleRadius;
  const int squareRadius  = radius * radius;
  const int excludeSide   = radius * 2;
  stCoord *sphere         = reserveMemory(sizeof(stCoord) * (radius * 2 + 1) * (radius * 2 + 1) * (radius * 2 + 1));
  stCoord *excludeVolume  = reserveMemory(sizeof(stCoord) * (excludeSide * 2 + 1) * (excludeSide * 2 + 1) * (excludeSide * 2 + 1));
  int sphereSize         = 0;
  int excludeVolumeSize  = 0;
  int length;

  // initialize of background molecular
  stCoord *stCoordPtr = sphere;
  for (int x = -radius; x <= radius; ++x) {
    for (int y = -radius; y <= radius; ++y) {
      for (int z = -radius; z <= radius; ++z) {
        length = x * x + y * y + z * z;
        if (length <= squareRadius) {
          (*stCoordPtr).x = x;
          (*stCoordPtr).y = y;
          (*stCoordPtr).z = z;
          ++stCoordPtr;
          ++sphereSize;
        }
      }
    }
  }

  // initialize of excludedVolume
  stCoordPtr = excludeVolume;
  for (int x = -excludeSide; x <= excludeSide; ++x) {
    for (int y = -excludeSide; y <= excludeSide; ++y) {
      for (int z = -excludeSide; z <= excludeSide; ++z) {
        for (int i = 0; i < sphereSize; ++i) {
          if (findElement(sphere, sphereSize, x + sphere[i].x, y + sphere[i].y, z + sphere[i].z) == true 
              && findElement(excludeVolume, excludeVolumeSize, x, y, z) == false) {
            (*stCoordPtr).x = x;
            (*stCoordPtr).y = y;
            (*stCoordPtr).z = z;
            ++stCoordPtr;
            ++excludeVolumeSize;
          }
        }
      }
    }
  }

  config.sphere               = sphere;
  config.excludeVolume        = excludeVolume;
  config.sphereSize           = sphereSize;
  config.excludeVolumeSize    = excludeVolumeSize;
  config.sphereNum            = (int)(config.obstacleNum / sphereSize) + ((config.obstacleNum % sphereSize == 0)? 0: 1);
}

bool configValidate ()
{
  if (config.latticeSize <= 0 ||
      config.volume <= config.obstacleNum) {
    return false;
  }

  return true;
}

void configDump()
{
  fprintf(stderr, "Lattice Size:              %d\n", config.latticeSize);
  fprintf(stderr, "Max Steps:                 %d\n", config.maxSteps);
  fprintf(stderr, "Obstacle Radius:           %d\n", config.obstacleRadius);
  fprintf(stderr, "Obstacle Rate:             %lf\n", config.obstacleRate);
  fprintf(stderr, "Obstacle Moving Interval:  %d\n", config.obstacleMovingInterval);
  fprintf(stderr, "Trial Number:              %d\n", config.trialNum);
  fprintf(stderr, "Sphere Size:               %d\n", config.sphereSize);
  fprintf(stderr, "Number Of Spheres:         %d\n", config.sphereNum);
  fprintf(stderr, "Percentage Of Spheres:     %lf\n", (double)(config.sphereNum * config.sphereSize) / (double)config.volume);
}

void freeConfig ()
{
  free(config.sphere);
  free(config.excludeVolume);
}

void usage ()
{
  fprintf(stderr, "Usage: [lattice size] [max step] [obstacle radius] [obstacle rate] [obstacle moving interval] [trial number]\n");
  exit(EXIT_FAILURE);
}


/*** random number generator ***/
gsl_rng* createRng (ulong seed)
{
  gsl_rng* rng;
  gsl_rng_env_setup();
  rng = gsl_rng_alloc(gsl_rng_default);
  gsl_rng_set(rng, seed);
  return rng;
}

void initRng ()
{
  ulong seed = time(NULL);

#ifdef _OPENMP
  rng = reserveMemory(sizeof(gsl_rng) * omp_get_max_threads());
  for (int i = 0; i < omp_get_max_threads(); ++i) {
    rng[i] = createRng(seed + i); 
  } 
#else
  rng = createRng(seed); 
#endif
}

static inline gsl_rng* getRandGen ()
{
  gsl_rng* randGen;

#ifdef _OPENMP
  randGen = rng[omp_get_thread_num()];
#else
  randGen = rng; 
#endif

  return randGen;
}

void freeRng()
{
#ifdef _OPENMP
  for (int i = 0; i < omp_get_max_threads(); ++i) {
    gsl_rng_free(rng[i]);
  }
  free(rng);
#else
  gsl_rng_free(rng);
#endif
}


/*** simulation ***/
/* initialize */
bool initSpace (bool *space, stCoord *obstacleCoord, int *orderList)
{
  // variables
  int volume, sphereSize, excludeVolumeSize, sphereNum, remainSphereNum, tmpNum, itReplace, *itOrderList, centerCoord[3];
  int tmpCoord[3];
  stCoord *sphere, *excludeVolume, *itObstacleCoord;
  gsl_rng *randGen;

  // initialize
  volume              = config.volume;
  sphereSize          = config.sphereSize;
  excludeVolumeSize   = config.excludeVolumeSize;
  sphereNum           = config.sphereNum;
  sphere              = config.sphere;
  excludeVolume       = config.excludeVolume;
  itObstacleCoord     = obstacleCoord;
  itOrderList         = orderList;
  randGen             = getRandGen();

  // rearrange order
  for (int i = volume - 1; i > 0; --i) {
    itReplace               = (i + 1) * gsl_rng_uniform(randGen);
    tmpNum                  = orderList[itReplace];
    orderList[itReplace]    = orderList[i];
    orderList[i]            = tmpNum;
  }

  // set obstacles
  remainSphereNum = sphereNum;
  memset(space, FIELD_FREE, sizeof(bool) * volume);
  while (remainSphereNum) {
    while (space[*itOrderList] == FIELD_OBSTACLE) {
      ++itOrderList;
      if (itOrderList == orderList + volume) return false;
    }

    indexToXyz(*itOrderList, centerCoord);
    (*itObstacleCoord).x  = centerCoord[0];
    (*itObstacleCoord).y  = centerCoord[1];
    (*itObstacleCoord).z  = centerCoord[2];

    for (int i = 0; i < excludeVolumeSize; ++i) {
      tmpCoord[0] = centerCoord[0] + excludeVolume[i].x;
      tmpCoord[1] = centerCoord[1] + excludeVolume[i].y;
      tmpCoord[2] = centerCoord[2] + excludeVolume[i].z;

      adoptPeriodicBoundaryCondition(tmpCoord);
      space[xyzToIndex((int*)tmpCoord)] = FIELD_OBSTACLE;
    }

    ++itObstacleCoord;
    --remainSphereNum;
  }

  memset(space, FIELD_FREE, sizeof(bool) * volume);
  for (int i = 0; i < sphereNum; ++i) {
    for (int j = 0; j < sphereSize; ++j) {
      tmpCoord[0] = obstacleCoord[i].x + sphere[j].x;
      tmpCoord[1] = obstacleCoord[i].y + sphere[j].y;
      tmpCoord[2] = obstacleCoord[i].z + sphere[j].z;

      adoptPeriodicBoundaryCondition(tmpCoord);
      space[xyzToIndex((int*)tmpCoord)] = FIELD_OBSTACLE;
    }
  }

  return true;
}

void updateObsatacle (bool *space, stCoord *obstacleCoord, int currentCoordId)
{
  const int volume               = config.volume;
  const int sphereNum            = config.sphereNum;
  const int excludeVolumeSize    = config.excludeVolumeSize;
  const int sphereSize           = config.sphereSize;
  stCoord *excludeVolume          = config.excludeVolume;
  stCoord *sphere                 = config.sphere;
  gsl_rng *randGen                = getRandGen();
  int itReplace, randNum, axis, centerCoord[3];
  int move, tmpCoord[3];
  stCoord replaceCoord;

  // rearrange obstacleCoord order
  for (int i = sphereNum - 1; i > 0; --i) {
    itReplace                   = (i + 1) * gsl_rng_uniform(randGen);
    //printf("itReplace: %d\n", itReplace);
    replaceCoord                = obstacleCoord[itReplace];
    obstacleCoord[itReplace]    = obstacleCoord[i];
    obstacleCoord[i]            = replaceCoord;
  }

  memset(space, FIELD_FREE, sizeof(bool) * volume);
  for (int i = 0; i < sphereNum; ++i) {
    randNum             = 6 * gsl_rng_uniform(randGen);
    move                = 2 * (randNum % 2) - 1;
    axis                = randNum / 2;
    centerCoord[0]      = obstacleCoord[i].x;
    centerCoord[1]      = obstacleCoord[i].y;
    centerCoord[2]      = obstacleCoord[i].z;
    centerCoord[axis]   = getNextCoord(centerCoord[axis] + move);

    if (space[xyzToIndex(centerCoord)] == FIELD_FREE && xyzToIndex(centerCoord) != currentCoordId) {
      switch (axis) {
        case 0:
          obstacleCoord[i].x  = centerCoord[0];
          break;
        case 1:
          obstacleCoord[i].y  = centerCoord[1];
          break;
        case 2:
          obstacleCoord[i].z  = centerCoord[2];
          break;
        default:
          break;
      }
    }

    for (int j = 0; j < excludeVolumeSize; ++j) {
      tmpCoord[0] = obstacleCoord[i].x + excludeVolume[j].x;
      tmpCoord[1] = obstacleCoord[i].y + excludeVolume[j].y;
      tmpCoord[2] = obstacleCoord[i].z + excludeVolume[j].z;

      adoptPeriodicBoundaryCondition(tmpCoord);
      space[xyzToIndex((int*)tmpCoord)] = FIELD_OBSTACLE;
    }
  }

  memset(space, FIELD_FREE, sizeof(bool) * volume);
  for (int i = 0; i < sphereNum; ++i) {
    for (int j = 0; j < sphereSize; ++j) {
      tmpCoord[0] = obstacleCoord[i].x + sphere[j].x;
      tmpCoord[1] = obstacleCoord[i].y + sphere[j].y;
      tmpCoord[2] = obstacleCoord[i].z + sphere[j].z;

      adoptPeriodicBoundaryCondition(tmpCoord);
      space[xyzToIndex((int*)tmpCoord)] = FIELD_OBSTACLE;
    }
  }
}



int initIndex (bool *space)
{
  int volume, index;
  gsl_rng *randGen;

  volume  = config.volume;
  randGen = getRandGen();

  while (1) {
    index = volume * gsl_rng_uniform(randGen);
    if (space[index]) break; 
  }

  return index;
}


void run (bool *space, stCoord *obstacleCoord, int *currentCoord, int *nextCoord, int* orderList, int *dist)
{
  // variables
  gsl_rng *randGen;
  int maxSteps, size, axis, randNum, obstacleMovingInterval, maxInitSpace, initSpaceCount;
  int move; 

  // initialize
  maxSteps                = config.maxSteps;
  size                    = config.latticeSize;
  obstacleMovingInterval  = config.obstacleMovingInterval;
  maxInitSpace            = 10; // ?
  initSpaceCount          = 0;
  randGen                 = getRandGen();
  while (++initSpaceCount <= maxInitSpace && !initSpace(space, obstacleCoord, orderList));// ?
  // while (!initSpace(space, obstacleCoord, orderList));

  // check if initializing space can be done
  if (initSpaceCount > maxInitSpace) {
    fprintf(stderr, "ERROR:initializing space cannot be done\n");    
    exit(EXIT_FAILURE);
  }
  indexToXyz(initIndex(space), currentCoord);

  // move molecules
  if (checkMobility(space, currentCoord, nextCoord)) {
    for (int i = 0; i < 3; ++i) {
      nextCoord[i] = currentCoord[i];
      dist[i] = 0;
    }

    for (int i = 0; i < maxSteps; ++i) {
      // update obstacle position
      if (i && i % obstacleMovingInterval == 0) {
        updateObsatacle (space, obstacleCoord, xyzToIndex(currentCoord));
      }

#pragma omp atomic
      results[i] += calMsd(dist);

      randNum = 6 * gsl_rng_uniform(randGen);
      axis = randNum / 2;
      move = 2 * (randNum % 2) - 1;

      nextCoord[axis] = getNextCoord(nextCoord[axis] + move);

      if (space[xyzToIndex(nextCoord)] == FIELD_FREE) {
        currentCoord[axis] = nextCoord[axis];
        dist[axis] += move;
      } else {
        nextCoord[axis] = currentCoord[axis];
      }
    } 
  }
}

void runSim ()
{
  int trialNum   = config.trialNum;
  initRng();

  // parallelization
#ifdef _OPENMP
  omp_set_num_threads(omp_get_max_threads());
#endif

#pragma omp parallel
  {
    bool *space             = reserveMemory(sizeof(bool) * config.volume);
    stCoord *obstacleCoord  = reserveMemory(sizeof(stCoord) * config.sphereNum);
    int *currentCoord      = reserveMemory(sizeof(int) * 3);
    int *nextCoord         = reserveMemory(sizeof(int) * 3);
    int *orderList         = reserveMemory(sizeof(int) * config.volume);
    int  *dist              = reserveMemory(sizeof(int)  * 3);

    for (int i = 0; i < config.volume; ++i) {
      orderList[i] = i;
    }

#pragma omp for
    for (int i = 0; i < trialNum; ++i) {
      fprintf(stderr, "Running Trial No.%d\n", i + 1);
      run (space, obstacleCoord, currentCoord, nextCoord, orderList, dist); 
    }

    free(space);
    free(obstacleCoord);
    free(currentCoord);
    free(nextCoord);
    free(orderList);
    free(dist);
  }

  freeRng();
}


/*** results ***/
void resultsInit()
{
  int maxSteps = config.maxSteps;
  results = reserveMemory(sizeof(ullong) * maxSteps);
  memset(results, 0, sizeof(ullong) * maxSteps);
}

void resultsDump()
{
  for (int i = 0; i < config.maxSteps; ++i) {
    fprintf(stdout, "%d %lf\n", i, (double)results[i] / (double)config.trialNum);
  }
}


/*** main function ***/
int main (int argc, char **argv)
{
  if (argc != 7) {
    usage();
  }

  // configurations
  fprintf(stderr, "### configurations ###\n");
  initConfig(argv);
  if (!configValidate()) usage();
  configDump();

  // simulation
  fprintf(stderr, "\n### simualtion started ###\n");
  resultsInit();
  runSim();

  // result
  fprintf(stderr, "\n### results ###\n");
  resultsDump();

  free(results);
  freeConfig();
  return 0;
}
